package org.ten60.netkernel.mod.pds;

import java.net.InetAddress;
import java.util.*;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFRequestReadOnly;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.nkf.INKFResponseReadOnly;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSNodeList;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.representation.impl.HDSPredicateFactory;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.urii.ISpace;
import org.netkernel.urii.ISpaceWithIdentity;


/****
 * Accessor for the pds:  and fpds: schemes - implements the absolute contextualization pattern
 * 
 * @author pjr
 *
 */
public class PDSAccessor extends StandardAccessorImpl
{	
	public static final String FPDS_METADATA_KEY="FPDS_METADATA_KEY";
	
	public PDSAccessor()
	{	declareThreadSafe();
		declareSupportedVerbs(INKFRequestReadOnly.VERB_DELETE|INKFRequestReadOnly.VERB_EXISTS|INKFRequestReadOnly.VERB_SOURCE|INKFRequestReadOnly.VERB_SINK);
	}
	
	public void onRequest(INKFRequestContext context) throws Exception
	{
		int verb=context.getThisRequest().getVerb();
		if (verb==INKFRequestReadOnly.VERB_META)
		{	super.onMeta(context);
		}
		else
		{	IHDSNode config=(IHDSNode)context.source("res:/etc/pdsConfig.xml", IHDSNode.class);
			String zone=(String)config.getFirstValue("/config/zone");
			
			String instance=null;
			instance=context.source("netkernel:/config/netkernel.instance.identifier", String.class);
			if(instance==null)
			{	//default to host name
				instance=InetAddress.getLocalHost().getCanonicalHostName();
			}
			boolean fallback=context.getThisRequest().getIdentifier().startsWith("fpds");
			String identifier=null;
			INKFResponse respOut;
			boolean frompds=false;
			if(fallback)
			{	identifier=context.getThisRequest().getIdentifier().substring(1);
			}
			else
			{	identifier=context.getThisRequest().getIdentifier();
			}
			
			context.setCWU(null);
			INKFRequest req=context.createRequest("active:pds");
			req.setVerb(verb);
			req.addArgument("pds", identifier);
			req.addArgument("zone", zone);
			req.addArgument("instance", instance);
			if(verb==INKFRequestReadOnly.VERB_SINK)
			{	req.addPrimaryArgumentFromResponse(context.getThisRequest().getPrimaryAsResponse());				
			}
			INKFResponseReadOnly respFromPDS=context.issueRequestForResponse(req);
			Object representation=respFromPDS.getRepresentation();
			
			if (verb==INKFRequestReadOnly.VERB_SOURCE)
			{	if (isIdentifierSetBoundary(identifier) && fallback)
				{	IHDSNode resSet = getResSet(identifier,context);
					Set<String> identifiers=new HashSet();
					IHDSNode pdsSet=(IHDSNode)representation;
					for (IHDSNode n:pdsSet.getNodes("/set/identifier"))
					{	identifiers.add("fpds"+n.getValue().toString().substring(3));
					}
					for (IHDSNode n:resSet.getNodes("/set/identifier"))
					{	identifiers.add("fpds"+n.getValue().toString().substring(3));
					}
					HDSBuilder b = new HDSBuilder();
					b.pushNode("set");
					for (String resId : identifiers)
					{	b.addNode("identifier",resId);
					}
					representation=b.getRoot();
					context.createResponseFrom(representation);
				}
				else if (representation==null)
				{	if (fallback)
					{	respFromPDS=context.sourceForResponse("res"+identifier.substring(3));
					
						//only need a golden thread on res resource, pds resource will already have one
						req=context.createRequest("active:attachGoldenThread");
						req.addArgument("id", PDSUtils.getGoldenThread(instance, zone, identifier));
						req.addArgument("id", PDSUtils.getGoldenThread(instance, zone, identifier.substring(0, identifier.lastIndexOf('/')+1)));
						context.issueRequest(req);
						respOut=context.createResponseFrom(respFromPDS);
					
						respOut.setHeader(FPDS_METADATA_KEY, "res");
					}
					else
					{	throw context.createFormattedException("EX_PDS_NOT_FOUND", "MSG_PDS_NOT_FOUND", null, null, identifier, zone);
					}
				}
				else
				{	respOut=context.createResponseFrom(respFromPDS);
					respOut.setHeader(FPDS_METADATA_KEY, "pds");
				}
			}
			else if (verb==INKFRequestReadOnly.VERB_EXISTS && fallback)
			{	boolean exists=(Boolean)representation;
				if (!exists)
				{	IHDSNode resSet = getResSet(getSetIdentifier(identifier),context);
					if (isIdentifierSetBoundary(identifier))
					{	exists=resSet.getFirstNode("/set/*")!=null;
					}
					else
					{	String expectingId="res"+identifier.substring(3);
						for (IHDSNode n: resSet.getNodes("/set/identifier"))
						{	if (n.getValue().toString().equals(expectingId))
							{	exists=true;
								break;
							}
						}
					}
				}
				context.createResponseFrom(exists);
			}
			else
			{	respOut=context.createResponseFrom(respFromPDS);
				frompds=true;
			}
		}
	}
	
	static boolean isIdentifierSetBoundary(String aIdentifier)
	{	return aIdentifier.endsWith("/");			
	}
	static String getSetIdentifier(String aIdentifier)
	{	return aIdentifier.substring(0, aIdentifier.lastIndexOf('/')+1);		
	}
	
	static IHDSNode getResSet(String aIdentifier, INKFRequestContext aContext) throws Exception
	{
		IRequestScopeLevel scope=aContext.getKernelContext().getThisKernelRequest().getRequestScope().getParent().getParent();
		ISpace space=scope.getSpace();
		while (scope!=null && (scope.isDurable() || !(space instanceof ISpaceWithIdentity)))
		{	scope=scope.getParent();
			space=scope.getSpace();
		}
		if (scope==null)
		{	throw new NKFException("no suitable space in request");
		}
			
		ISpaceWithIdentity swi=(ISpaceWithIdentity)space;
		INKFRequest req=aContext.createRequest("active:sls");
		req.addArgumentAsAbsolute("space", swi.getIdentifier().toString());
		//req.addArgumentAsAbsolute("space", "urn:org:netkernel:tutorial:gettingstarted");
		req.addArgumentAsAbsolute("version", swi.getVersion().toString());
		req.setRepresentationClass(IHDSNode.class);
		IHDSNode result=(IHDSNode)aContext.issueRequest(req);
		String id="res"+aIdentifier.substring(3);
		IHDSNode dirNode=result.getNodes("//dir").filter(HDSPredicateFactory.namedChildStringEquals("@id", id)).getFirstNode();
		
		HDSBuilder b = new HDSBuilder();
		b.pushNode("set");
		if (dirNode!=null)
		{	IHDSNodeList ress=dirNode.getNodes("res/@id");
			for (IHDSNode res: ress)
			{	String resId=res.getValue().toString();
				b.addNode("identifier",resId);
			}
			if (ress.size()==0)
			{	b.addNode("empty",null);
			}
		}
		return b.getRoot();
	}
	
}
